#region Using Statements
using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Storage;
#endregion

namespace FallingBlockPuzzle {
	public enum Type : byte {
		Empty,
		Wall,
	}
	public class Palette {
		readonly Color[] colors = {
			Color .TransparentBlack, Color.Black, Color.IndianRed, Color.ForestGreen, Color.SkyBlue, Color.Khaki };
		public Color this[Type type] { get { return colors[(int)type]; } }
	}

	public class PiecePattern {
		readonly Type[] pattern;

		public PiecePattern(Type[] pattern) {
			this.pattern = pattern;
		}

		public int Length { get { return pattern.Length; } }
		public Type this[int index] { get { return pattern[index]; } }
	}
	public class Piece : Blocks {
		readonly Blocks field;
		readonly FieldRotator rotator;
		bool enabled = true;

		/// <summary>
		/// P̃ubN̕s[XB
		/// </summary>
		/// <param name="block"></param>
		public Piece(Blocks field, Block block)
			: base(1, 1) {
			this.field = field;
			this[Point.Zero].Offset = block.Offset;
			this[Point.Zero].Type = block.Type;
			type = block.Type;
		}

		public Piece(Blocks field)
			: base(4, 4) {
			this.field = field;
			rotator = new FieldRotator(this);
		}

		public bool Enabled { get { return enabled; } set { enabled = value; } }

		public IEnumerable<Point> BlockPositions {
			get {
				Point position;
				for (position.Y = 0; position.Y < Height; position.Y++) {
					for (position.X = 0; position.X < Width; position.X++) {
						Block block = this[position];
						if (block.Type == type) {
							yield return block.Position;
						}
					}
				}
			}
		}

		public void Initialize(PiecePattern pattern, Type type) {
			this.type = type;
			Point position;
			int i = 0;
			for (position.Y = 0; position.Y < Height; position.Y++) {
				for (position.X = 0; position.X < Width; position.X++) {
					Block block = this[position];
					block.Type = pattern[i++] == Type.Empty ? Type.Empty : type;
					block.Offset.X = position.X * BlockVisual.Width;
					block.Offset.Y = position.Y * BlockVisual.Height;
				}
			}

		}
		public void CopyTo(Piece other) {
			other.type = type;
			Point position;
			for (position.Y = 0; position.Y < Height; position.Y++) {
				for (position.X = 0; position.X < Width; position.X++) {
					Block block = other[position];
					block.Type = this[position].Type;
					block.Offset.X = position.X * BlockVisual.Width;
					block.Offset.Y = position.Y * BlockVisual.Height;
				}
			}
			other.position.X = 0;
			other.position.Y = 0;
		}

		public void Rewind() {
			position.Y--;
		}
		static readonly Vector2 Gravity = Vector2.UnitY;
		static readonly Vector2 FallSpeed = Gravity * 0.5f;
		public void Translate(Vector2 translation) {
			Point position;
			for (position.Y = 0; position.Y < Height; position.Y++) {
				for (position.X = 0; position.X < Width; position.X++) {
					Block block = this[position];
					block.Offset += translation;
				}
			}
			this.position.X = (int)Math.Floor((this[Point.Zero].Offset.X + 0.5f) / BlockVisual.Width);
			this.position.Y = (int)Math.Floor((this[Point.Zero].Offset.Y + BlockVisual.Height - 1) / BlockVisual.Height);
		}
		public void Update() {
			Translate(FallSpeed);
		}
		/// <summary>
		/// s[XtB[hɓ\tB
		/// </summary>
		public bool TryPaste() {
			bool completed = true;
			Point start = new Point(this.position.X, this.position.Y);
			Point end = new Point(Width + start.X, Height + start.Y);
			Point position;
			Point global;
			for (position.Y = 0; position.Y < Height; position.Y++) {
				global.Y = position.Y + this.position.Y;
				for (position.X = 0; position.X < Width; position.X++) {
					global.X = position.X + this.position.X;
					Block source = this[position];
					if (source.Type != Type.Empty) {
						if (!field.Contains(global))
							completed = false;
						Block destination = field[global];
						destination.Type = source.Type;
					}
				}
			}
			return completed;
		}
		Point position;
		private Type type;

		public bool HitTest() {
			Point local;
			Point global;
			for (local.Y = 0; local.Y < Height; local.Y++) {
				global.Y = local.Y + this.position.Y;
				for (local.X = 0; local.X < Width; local.X++) {
					global.X = local.X + this.position.X;
					if (this[local].Type != Type.Empty
						&& field[global].Type != Type.Empty) {
						return true;
					}
				}
			}
			return false;
		}

		public Type Type {
			get { return type; }
			set { type = value; }
		}

		public void TranslateLeft() {
			Translate(new Vector2(-BlockVisual.Width, 0));
			if (HitTest()) {
				Translate(new Vector2(+BlockVisual.Width, 0));
			}
		}
		public void TranslateRight() {
			Translate(new Vector2(+BlockVisual.Width, 0));
			if (HitTest()) {
				Translate(new Vector2(-BlockVisual.Width, 0));
			}
		}

		public void RotateLeft() {
			rotator.RotateLeft();
			if (HitTest()) {
				rotator.RotateRight();
			}
		}
		public void RotateRight() {
			rotator.RotateRight();
			if (HitTest()) {
				rotator.RotateLeft();
			}
		}
		public Point Position { get { return position; } set { position = value; } }
	}
	public class Block {
		public static readonly Block Wall = new Block(Type.Wall);
		public static readonly Block Empty = new Block(Type.Empty);
		public Vector2 Offset;
		private Type type;
		private bool marked;

		public Block() { }
		public Block(Type type) {
			this.type = type;
		}
		public override string ToString() {
			return string.Format("({0},{1}){2}{3}", Offset.X, Offset.Y, marked ? "*" : ":", type);
		}
		public bool Marked {
			get { return marked; }
			set { marked = value; }
		}

		public Type Type {
			get { return type; }
			set { type = value; }
		}
		public bool TryMark() { if (marked) return false; return marked = true; }
		public bool TryUnmark() { if (!marked) return false; marked = false; return true; }

		public Point Position {
			get {
				Point value;
				value.X = (int)Math.Floor((Offset.X + 0.5f) / BlockVisual.Width);
				value.Y = (int)Math.Floor((Offset.Y + 0.5f) / BlockVisual.Height);
				return value;
			}
		}
	}
	public class FieldRotator {
		protected readonly Block[] copy;
		protected readonly Blocks blocks;
		public FieldRotator(Blocks blocks) {
			if (blocks.Width != blocks.Height)
				throw new ArgumentException("cTCŶ̂ɑΉĂ܂B");
			copy = new Block[blocks.Width * blocks.Height];
			this.blocks = blocks;
		}
		public void RotateLeft() {
			blocks.Data.CopyTo(copy, 0);
			Vector2 offset = blocks[Point.Zero].Offset;
			Point position;
			int i = 0;
			for (position.X = blocks.Width; --position.X >= 0; ) {
				for (position.Y = 0; position.Y < blocks.Height; position.Y++) {
					blocks[position] = copy[i++];
					blocks[position].Offset.X = position.X * BlockVisual.Width + offset.X;
					blocks[position].Offset.Y = position.Y * BlockVisual.Height + offset.Y;
				}
			}
		}
		public void RotateRight() {
			blocks.Data.CopyTo(copy, 0);
			Vector2 offset = blocks[Point.Zero].Offset;
			Point position;
			int i = 0;
			for (position.X = 0; position.X < blocks.Width; position.X++) {
				for (position.Y = blocks.Height; --position.Y >= 0; ) {
					blocks[position] = copy[i++];
					blocks[position].Offset.X = position.X * BlockVisual.Width + offset.X;
					blocks[position].Offset.Y = position.Y * BlockVisual.Height + offset.Y;
				}
			}
		}
	}

	public class Blocks {
		public readonly int Width;
		public readonly int Height;
		protected readonly Block[] blocks;
		public Block[] Data { get { return blocks; } }
		public Blocks(int width, int height) {
			Width = width;
			Height = height;
			blocks = new Block[Width * Height];
			Point position;
			int i = 0;
			for (position.Y = 0; position.Y < Height; position.Y++) {
				for (position.X = 0; position.X < Width; position.X++) {
					Block block = blocks[i++] = new Block();
					block.Offset = new Vector2(position.X * BlockVisual.Width, position.Y * BlockVisual.Height);
				}
			}
		}
		public bool Contains(Point position) {
			if (position.X < 0 || position.Y < 0 || position.X >= Width || position.Y >= Height)
				return false;
			return true;
		}
		public Block this[Point position] {
			get {
				if (position.X < 0 || position.X >= Width || position.Y >= Height)
					return Block.Wall;
				if (position.Y < 0)
					return Block.Empty;
				return blocks[position.X + position.Y * Width];
			}
			set {
				blocks[position.X + position.Y * Width] = value;
			}
		}
		public void Unmark() {
			for (int i = 0; i < blocks.Length; i++) {
				blocks[i].TryUnmark();
			}
		}
		public IEnumerable<Block> UnmarkMates(Type type) {
			for (int i = 0; i < blocks.Length; i++) {
				if (blocks[i].TryUnmark() && blocks[i].Type == type)
					yield return blocks[i];
			}
		}
		public IEnumerable<Block> UnmarkMates() {
			for (int i = 0; i < blocks.Length; i++) {
				if (blocks[i].TryUnmark())
					yield return blocks[i];
			}
		}
	}
}
