#region Using Statements
using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Storage;
#endregion

namespace FallingBlockPuzzle {

	public class BlockStack {
		readonly Blocks blocks;
		public BlockStack(Blocks blocks) {
			this.blocks = blocks;
		}
		public IEnumerable<Block> MarkMates(Point position, Predicate<Block> match) {
			for (; position.Y >= 0; position.Y--) {
				Block block = blocks[position];
				if (block.Marked || !match(block))
					yield break;
				block.Marked = true;
				yield return block;
			}
		}
	}
	public class BlockRegion {
		readonly Blocks blocks;
		public BlockRegion(Blocks blocks) {
			this.blocks = blocks;
		}

		public IEnumerable<Block> MarkMates(Point position, Predicate<Block> match) {
			Block block = blocks[position];
			if (block.Marked || !match(block))
				yield break;
			block.Marked = true;
			yield return block;
			position.X--;
			foreach (Block neighbor in MarkMates(position, match))
				yield return neighbor;
			position.X += 2;
			foreach (Block neighbor in MarkMates(position, match))
				yield return neighbor;
			position.X--;
			position.Y--;
			foreach (Block neighbor in MarkMates(position, match))
				yield return neighbor;
			position.Y += 2;
			foreach (Block neighbor in MarkMates(position, match))
				yield return neighbor;
		}

	}
	public class Randomizer {
		Blocks blocks;
		Random random = new Random();
		public Randomizer(Blocks blocks) {
			this.blocks = blocks;
		}
		public void Randomize(int numberTypes) {
			Point position;
			for (position.Y = 0; position.Y < blocks.Height; position.Y++) {
				for (position.X = 0; position.X < blocks.Width; position.X++) {
					blocks[position].Type = (Type)random.Next(numberTypes);
				}
			}
		}
	}
	public class StageState {
		public int NumberTypes = 4;
		public int VanishingRegionSize = 4;
		public Random Random = new Random();
	}
	public class PieceStream {
		static readonly PiecePattern[] patterns = {
			new PiecePattern(new Type[] {
				Type.Empty, Type.Empty, Type.Empty, Type.Empty,
				Type.Empty, Type.Wall, Type.Empty, Type.Empty,
				Type.Empty, Type.Wall, Type.Empty, Type.Empty,
				Type.Empty, Type.Empty, Type.Empty, Type.Empty}),
		};

		readonly Blocks field;
		readonly Piece current;
		readonly Piece next;
		readonly List<Piece> freefalls = new List<Piece>();
		readonly StageState state;
		int speed = 1;

		public Piece Current { get { return current; } }
		public Piece Next { get { return next; } }
		public List<Piece> Freefalls { get { return freefalls; } }
		public int Speed { get { return speed; } set { speed = value; } }

		public PieceStream(StageState state, Blocks field) {
			this.state = state;
			this.field = field;
			current = new Piece(field);
			next = new Piece(field);
			next.Initialize(
				patterns[state.Random.Next(patterns.Length)],
				(Type)state.Random.Next(state.NumberTypes) + 2);
			Spawn();
		}
		public void Spawn() {
			next.CopyTo(current);
			next.Initialize(
				patterns[state.Random.Next(patterns.Length)],
				(Type)state.Random.Next(state.NumberTypes) + 2);
			next.Translate(new Vector2(32 * BlockVisual.Width, 0));
			current.Translate(new Vector2(
				((field.Width - current.Width) / 2) * BlockVisual.Width,
				-current.Height * BlockVisual.Height));
			current.Enabled = true;
		}
	}
	public class PieceInput : GameComponent {
		readonly Blocks field;
		KeyboardState current;
		KeyboardState previous;
		readonly PieceStream pieces;

		public PieceStream Pieces { get { return pieces; } }
		public PieceInput(Game game, StageState state, Blocks field, PieceStream pieces)
			: base(game) {
			this.field = field; this.pieces = pieces;
		}
		public override void Update(GameTime gameTime) {
			previous = current;
			current = Keyboard.GetState();

			if (pieces.Current.Enabled) {
				if (current.IsKeyDown(Keys.Left) && previous.IsKeyUp(Keys.Left)) {
					pieces.Current.TranslateLeft();
				} else if (current.IsKeyDown(Keys.Right) && previous.IsKeyUp(Keys.Right)) {
					pieces.Current.TranslateRight();
				} else if (current.IsKeyDown(Keys.Z) && previous.IsKeyUp(Keys.Z)) {
					pieces.Current.RotateLeft();
				} else if (current.IsKeyDown(Keys.X) && previous.IsKeyUp(Keys.X)) {
					pieces.Current.RotateRight();
				}
				if (current.IsKeyDown(Keys.Down)) {
					pieces.Speed = 16;
				} else {
					pieces.Speed = 1;
				}
			}
			base.Update(gameTime);
		}
	}
	public class PieceControl : GameComponent {
		readonly PieceStream pieces;
		readonly Blocks field;
		readonly StageState state;
		readonly BlockRegion region;
		readonly BlockStack stack;

		public PieceStream Pieces { get { return pieces; } }

		public PieceControl(Game game, StageState state, Blocks field)
			: base(game) {
			this.field = field;
			this.state = state;
			region = new BlockRegion(field);
			stack = new BlockStack(field);
			pieces = new PieceStream(state, field);
		}

		readonly List<Block> selectedBlocks = new List<Block>();
		readonly List<Block> selections = new List<Block>();
		/// <summary>
		/// s[X̗i߂B
		/// </summary>
		/// <param name="piece"></param>
		/// <returns>n̂Ƃtrue</returns>
		public bool Fall(Piece piece) {
			for (int i = 0; i < pieces.Speed; i++) {
				piece.Update();
				if (piece.HitTest()) {
					piece.Rewind();
					if (!piece.TryPaste()) {
						overflow = true;
						return true;
					}

					foreach (Point position in piece.BlockPositions) {
						//Console.WriteLine("{0} {1}", position, field[position]);
						selections.AddRange(region.MarkMates(position, delegate(Block block) {
							return block.Type == piece.Type;
						}));
					}
					return true;
				}
			}
			return false;
		}
		public bool TryVanish() {
			if (selections.Count >= state.VanishingRegionSize) {
				selectedBlocks.AddRange(selections);
				for (int i = 0; i < selections.Count; i++) {
					selections[i].TryUnmark();
					selections[i].Type = Type.Empty;
				}
				selections.Clear();
				return true;
			} else {
				for (int i = 0; i < selections.Count; i++) {
					selections[i].TryUnmark();
				}
				selections.Clear();
				return false;
			}
		}
		public bool TryChain() {
			bool fall = false;

			for (int i = 0; i < selectedBlocks.Count; i++) {
				Block block = selectedBlocks[i];
				Point position = block.Position;
				position.Y--;
				foreach (Block falling in stack.MarkMates(position, Droppable)) {
					Piece piece = new Piece(field, falling);
					pieces.Freefalls.Add(piece);
					falling.Type = Type.Empty;
					fall = true;
				}
			}
			field.Unmark();

			selectedBlocks.Clear();
			return fall;
		}

		bool overflow;

		public bool Overflow { get { return overflow; } }

		public override void Update(GameTime gameTime) {

			if (pieces.Freefalls.Count > 0) {
				for (int i = 0; i < pieces.Freefalls.Count; ) {
					if (Fall(pieces.Freefalls[i])) {
						pieces.Freefalls.RemoveAt(i);
						TryVanish();
					} else
						i++;
				}
				if (pieces.Freefalls.Count == 0) {
					if (!TryChain())
						pieces.Spawn();
				}
			} else {
				pieces.Current.Enabled = true;
				if (Fall(pieces.Current)) {
					pieces.Current.Enabled = false;
					pieces.Speed = 4;
					if (!TryVanish() || !TryChain())
						pieces.Spawn();
				}
			}

			/*
			foreach (Block block in field.UnmarkMates(pieces.Current.Type)) {
				block.Type = Type.Empty;
			}*/

			base.Update(gameTime);
		}

		bool Droppable(Block block) {
			return block.Type != Type.Empty && block.Type != Type.Wall;
		}
	}

}
